/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "wchunk.hh"
#include "wcondition.hh"
#include "weventlist.hh"
#include "woutput.hh"
#include "wparameters.hh"
#include "wtile.hh"
#include "wtransform.hh"
#include "matlab_fcs.hh"
#include "MultiDacc.hh"
#include "lcl_array.hh"
#include "usertime.hh"
#include <iostream>
#include <iomanip>

using namespace std;

// bugs fixed: Injection time shift and data factors incorrect if data 
// bounced by incorrect state.

namespace wpipe {
   //=====================================  Contruct a chunker
   wchunk::wchunk(void) 
      : startTime(0), currentTime(0), blockNo(0)
   {}

   //=====================================  Process a block of data.
   void
   wchunk::process(double blockDuration, double blockOverlap, 
		   const wtile&  tiling, const wparameters& pars, 
		   MultiDacc& frameSource, woutput& outputFiles, 
		   int debugLevel, str_vect& channelNames) 
   {
      user_timer cpuTime;

      /////////////////////////////////////////////////////////////////////////
      //                               read parameters                       //
      /////////////////////////////////////////////////////////////////////////
      // extract parameters from pars structure
      channelNames = pars.channelNames;
      str_vect frameTypes = pars.frameTypes;

      double fSample = pars.sampleFrequency;
      dble_vect timeShifts = pars.timeShifts;
      str_vect  injectionNames = pars.injectionNames;
      str_vect  injectionTypes = pars.injectionTypes;
      dble_vect injectionTimeShifts = pars.injectionTimeShifts;
      size_t numberOfChannels = pars.numberOfChannels();

      //--------------------------------  Reset inputs for new channel list
      if (numberOfChannels != input_data.size()) {
	 currentTime = Time(0);
	 startTime = Time(0);
	 input_data.clear();
	 input_data.resize(numberOfChannels);
      }

      /////////////////////////////////////////////////////////////////////////
      //                             determine times                         //
      /////////////////////////////////////////////////////////////////////////

      double fillDuration = !currentTime ? blockDuration 
	                                 : blockDuration-blockOverlap;

      int rc = frameSource.fillData(fillDuration);
      if (rc) return;
	 
      //-------------------------------  Append new data.
      Time fillStartTime = frameSource.getFillTime();
      if (fillStartTime == currentTime) {
	 Time newStart = currentTime - blockOverlap;
	 Interval tDelete = newStart - startTime;
	 for (size_t i=0; i<numberOfChannels; i++) {
	    if (tDelete > Interval(0.0)) input_data[i].eraseStart(tDelete);
	    input_data[i].Append(*frameSource.refData(channelNames[i].c_str()));
	 }
	 startTime = newStart;
      }

      //-------------------------------  No overlap data.
      else {
	 for (size_t i=0; i<numberOfChannels; i++) {
	    TSeries* ts = frameSource.refData(channelNames[i].c_str());
	    if (!ts) {
	       cerr << "Data not available for channel: " << channelNames[i]
		    << endl;
	       frameSource.list(cerr);
	       throw runtime_error("wchunk: Error reading data");
	    }
	    input_data[i] = *ts;
	 }
	 startTime = fillStartTime;
      }
      currentTime = frameSource.getCurrentTime();
      if (currentTime < startTime + blockDuration) return;

      // stop time of block
      // -- assume start time is aligned for now.
      Time blockStartTime = startTime;
      Time blockStopTime  = currentTime;

      // report status
      if (debugLevel >= 1) {
	 cout << fixed << setprecision(2) 
	      << "  chunk start time:        " << blockStartTime.totalS() 
	      << endl
	      << "  chunk stop time:         " << blockStopTime.totalS()
	      << endl;
      }

      /////////////////////////////////////////////////////////////////////////
      //                       transform sky coordinate                      //
      /////////////////////////////////////////////////////////////////////////
      dble_vect coordinate;

#ifndef NO_COHERENT
      // if (the skyPosition parameter is not empty, determine coordinate
      if (!skyPosition.empty()) {

	 // gps center time of block
	 blockCenterTime = blockStartTime + (pars.blockDuration / 2);
	 if (debugLevel >= 2) {
	    cout << "  block center time:       " << blockCenterTime << endl;
	 }

	 // convert sky coordinate into geocentric coordinate
	 wconvertskycoordinates coordinate(skyPosition, blockCenterTime, 
					   skyCoordinateSystem, "geocentric");

	 if (debugLevel >= 1) {
	    cout <<  "  sky coordinate:          [" << coordinate[1] << " " 
		 << coordinate[2] << endl;
	 }
      }
#endif

      /////////////////////////////////////////////////////////////////////////
      //                             read detector state                     //
      /////////////////////////////////////////////////////////////////////////
      vector<int>  chanIdent;

      // if (detector state channels are requested
      if (!pars.stateNames.empty()) {

	 // report status
	 wlog(debugLevel, 1, "  reading detector state");

	 // read detector data
	 size_t nStateChans = pars.stateNames.size();
	 tser_vect stateData(nStateChans);
	 for (size_t i=0; i<nStateChans; i++) {
	    stateData[i] = *frameSource.refData(pars.stateNames[i].c_str());
	 }

	 // bool array for tracking valid channels
	 vector<bool> validChannels(numberOfChannels, true);
  
	 // loop over channels
	 for (size_t chanNumber=0; chanNumber<numberOfChannels; chanNumber++) {

	    // check for missing data
	    if (stateData[chanNumber].empty()) {
	       if (debugLevel >= 1) {
		  cout << "    " << pars.stateNames[chanNumber] 
		       << ":        state read error\n" << endl;
	       }
	       if (pars.errorOnStateError) {
		  error("error reading detector state");
	       } else {
		  validChannels[chanNumber] = false;
		  continue;
	       }
	    }

	    // loop over data words
	    int mask = int(pars.stateMasks[chanNumber]);
	    int N    = stateData[chanNumber].getNSample();
	    lcl_array<int> state_chan(N);
	    stateData[chanNumber].getData(N, state_chan);
	    for (int i=0; i<N; i++) {
	       int xi = state_chan[i];

	       // check for negative data
	       if (xi < 0) {
		  if (debugLevel >= 1) {
		     cout <<  "    " << pars.stateNames[chanNumber] 
			  << ":        state data negative" << endl; 
		  }
		  if (pars.errorOnStateError) {
		     error("error reading detector state");
		  } else {
		     validChannels[chanNumber] = false;
		     break;
		  }
	       }

	       //  check state
	       if ((xi & mask) != mask) {
		  if (debugLevel >= 1) {
		     cout <<  "    " << pars.stateNames[chanNumber] 
			  << ":        detector not in requested state\n" 
			  << endl; 
		  }
		  validChannels[chanNumber] = false;
		  break;
	       }
	    }
	 }

	 //-----------------------------------  Remove channels failing checks
	 numberOfChannels = 0;
	 for (size_t i=0; i<validChannels.size(); i++) {
	    if (!validChannels[i]) continue;
	    chanIdent.push_back(i);
	    if (i != numberOfChannels) {
	       channelNames[numberOfChannels]   = channelNames[i];
	       frameTypes[numberOfChannels]     = frameTypes[i];
	       timeShifts[numberOfChannels]     = timeShifts[i];
	       injectionNames[numberOfChannels] = injectionNames[i];
	       injectionTypes[numberOfChannels] = injectionTypes[i];
	       injectionTimeShifts[numberOfChannels] = injectionTimeShifts[i];
	    }
	    numberOfChannels++;
	 }

	 //----------------------------------- Erase unavailable channels.
	 channelNames.erase(channelNames.begin()+numberOfChannels,
			    channelNames.end());
	 frameTypes.erase(frameTypes.begin()+numberOfChannels, 
			  frameTypes.end());
	 timeShifts.erase(timeShifts.begin()+numberOfChannels, 
			  timeShifts.end());
	 injectionNames.erase(injectionNames.begin()+numberOfChannels,
			      injectionNames.end());
	 injectionTypes.erase(injectionTypes.begin()+numberOfChannels,
			      injectionTypes.end());
	 injectionTimeShifts.erase(injectionTimeShifts.begin()
				   +numberOfChannels,
				   injectionTimeShifts.end());

      }    // end test for requested detector state

      // error if (there is no data after state checks
      if (channelNames.empty()) {
	 wlog(debugLevel, 1, "    no valid data to analyze");
      }
  
      /////////////////////////////////////////////////////////////////////////
      //                         read detector data                          //
      /////////////////////////////////////////////////////////////////////////
      else {
	 wlog(debugLevel, 1, "  reading detector data");

	 // read detector data
  
	 //-------------- pre-process data
	 tser_vect data(numberOfChannels);
	 string networkString;
	 for (size_t chan=0; chan < numberOfChannels; chan++) {
	    data[chan]=input_data[chan].extract(blockStartTime, blockDuration);
	    if (debugLevel) {
	       cerr << "Channel " << channelNames[chan] << " start: " 
		    << data[chan].getStartTime() << " end: "
		    << data[chan].getEndTime() << endl;
	    }
	    int chanID = chanIdent[chan];

	    // test for missing detector data
	    if (data[chan].empty()) error("error reading detector data");

	    // standardize channel names
	    channelNames[chan]=strrep(channelNames[chan], ";", ":");

	    // apply multiplicative factor to data if (specified
	    if (pars.dataFactors[chanID] != 1.0) {
	       data[chan] *= pars.dataFactors[chan];
	    }

	    // report data loaded
	    if (debugLevel >= 1) {
	       cout <<  "    " << channelNames[chan] << ":        loaded x "
		    <<  pars.dataFactors[chanID] << endl;
	    }

	    // record analyzed network
	    if (chan != 0) networkString += ",";
	    networkString += channelNames[chan].substr(0,2);
	 }

	 //////////////////////////////////////////////////////////////////////
	 //                       read injection data                        //
	 //////////////////////////////////////////////////////////////////////

	 // if (any injections are requested
	 if (pars.injectionChannels() != 0) {

	    // report status
	    wlog(debugLevel, 1, "  reading injection data");

	    // read injection data
	    tser_vect injectionData;

	    // test for missing or inconsistent injection data
	    for (size_t chanNumber=0; chanNumber<injectionData.size(); 
		 chanNumber++) {
	       int chanID = chanIdent[chanNumber];

	       if (injectionData[chanNumber].empty()) {
		  error("error reading injection data");
	       }

	       if (injectionData[chanNumber].getTStep() != 
		   data[chanNumber].getTStep()) {
		  error("inconsistent detector and injection data");
	       }

	       // Scale injection data and add it in.
	       injectionData[chanNumber] *= pars.injectionFactors[chanID];
	       data[chanNumber] += injectionData[chanNumber];

	       // report injection data loaded
	       if (debugLevel >= 1) {
		  cout <<  "    " << injectionNames[chanNumber] 
		       << ":        loaded x " << pars.injectionFactors[chanID] 
		       << endl;
	       }
	    }
	 }  // end test for requested injections

	 //////////////////////////////////////////////////////////////////////
	 //                          resample data                           //
	 //////////////////////////////////////////////////////////////////////

	 // report status
	 wlog(debugLevel, 1, "  resampling data");

	 // resample data
	 tser_vect resdata = resample.wresample(data, fSample);
	 if (debugLevel) {
	    cerr << "Resampled channel " << channelNames[0] << " start: " 
		 << resdata[0].getStartTime() << " end: "
		 << resdata[0].getEndTime() << endl;
	 }

	 //////////////////////////////////////////////////////////////////////
	 //                          condition data                          //
	 //////////////////////////////////////////////////////////////////////
	 wlog(debugLevel, 1, "  conditioning data");

	 // condition data
	 wcondition conditioner(resdata, tiling, pars.doubleWhiten);

	 //////////////////////////////////////////////////////////////////////
	 //                      transform and threshold                     //
	 //////////////////////////////////////////////////////////////////////

	 // setup thresholds ranges
	 Time thresholdReferenceTime = blockStartTime + pars.blockDuration*0.5;
	 dble_vect thresholdTimeRange;
	 dble_vect thresholdFrequencyRange;
	 dble_vect thresholdQRange;

	 // The original version combined the transformation and triggering.
	 // I don't understand why this would be necessary except for memory 
	 // management. this could be done in c++ but it certainly isn't a 
	 // critical problem. If it should prove to be important at a later 
	 // time, the channel-by-channel transform+thresholding can be 
	 // implemented using the single channel transform and thresholding 
	 // classes.
	 wlog(debugLevel, 1, "  transforming data");
	 dft_vect condDft, coeffs;
	 conditioner.whitenedDFT(condDft);
	 conditioner.coefficientDFT(coeffs);

	 wtransform transforms(condDft, tiling, pars.outlierFactor, 
			       pars.analysisMode, channelNames, coeffs, 
			       coordinate);

	 wlog(debugLevel, 1, "  thresholding data");
	 weventstack triggers, triggersThresholded, triggersDownselected;
	 triggersThresholded.wthreshold(transforms, tiling, blockStartTime, 
					pars.falseEventRate, 
					thresholdReferenceTime, 
					thresholdTimeRange, 
					thresholdFrequencyRange, 
					thresholdQRange, 
					pars.maximumSignificants, 
					pars.analysisMode, pars.falseVetoRate, 
					pars.uncertaintyFactor, 
					pars.correlationFactor, debugLevel);

	 // report significant tile numbers
	 if (debugLevel >= 1) {
	    // triggersThresholded[0][0].dump(cout);
	    triggersThresholded.status(cout);
	 }

	 //////////////////////////////////////////////////////////////////////
	 //                     select significant triggers                  //
	 //////////////////////////////////////////////////////////////////////

	 // report status
	 wlog(debugLevel, 1, "  selecting significant triggers");

	 // excluding overlapping significant tiles
	 triggersDownselected.wselect(triggersThresholded, 
				      pars.durationInflation,
				      pars.bandwidthInflation, 
				      pars.maximumTriggers, 
				      debugLevel);

	 // replace triggers
	 triggers = triggersDownselected;

	 // report downselected tile numbers
	 if (debugLevel >= 1) {
	    // triggers[0][0].dump(cout);
	    triggers.status(cout);
	 }

	 //////////////////////////////////////////////////////////////////////
	 //                    write single detector triggers                //
	 //////////////////////////////////////////////////////////////////////

	 if (outputFiles.enabled("DOWNSELECT")) {
      
	    // report status
	    wlog(debugLevel, 1, "  writing selected triggers");

	    // write triggers
	    triggers.writeEvents(outputFiles["DOWNSELECT"], pars.triggerFields, 
				 pars.triggerFormat);
	 }

	 //////////////////////////////////////////////////////////////////////
	 //                     veto inconsistent triggers                   //
	 //////////////////////////////////////////////////////////////////////
#ifndef NO_COHERENT
	 // if veto is requested and analysis mode is coherent
	 if (pars.applyVeto && pars.analysisMode == "coherent") {
	    wlog(debugLevel, 1, "  applying veto");

	    // apply null stream veto
	    eventstack triggersVetoed;
	    triggersVetoed.wveto(triggers, pars.durationInflation, 
				 pars.bandwidthInflation, 
				 vetoDurationFactor, vetoBandwidthFactor, 
				 maximumConsistents, debugLevel);
  
	    // replace triggers
	    triggers = triggersVetoed;

	    // report consistent tile numbers
	    if (debugLevel >= 1) {
	       triggersVetoed.status(cout);
	    }
	 } // end test for apply veto
#endif // !defined(NO_COHERENT)

	 //////////////////////////////////////////////////////////////////////
	 //                         cluster triggers                         //
	 //////////////////////////////////////////////////////////////////////

	 // if (clustering is requested,
	 if (pars.applyClustering || pars.bayesian) {
	    wlog(debugLevel, 1, "  applying clustering");

	    // apply clusterings
	    weventstack triggersClustered = triggers;
	    weventstack clusters;

	    //-------------------------------  Density clustering
	    string method = toupper(pars.clusterMethod);
	    if (pars.clusterMethod == "density") {
	       clusters.wcluster(triggersClustered, pars.clusterRadius, 
				 pars.clusterDensity, pars.clusterSingles, 
				 pars.distanceMetric, pars.durationInflation, 
				 pars.bandwidthInflation, debugLevel);
	    }

	    //-------------------------------  Hierarchical clustering
	    else if (pars.clusterMethod == "hierarchical") {
	       clusters.wcluster(triggersClustered, pars.clusterLinkage, 
				 pars.clusterCriterion, pars.clusterThreshold, 
				 pars.distanceMetric, pars.durationInflation, 
				 pars.bandwidthInflation, debugLevel);
	    }

	    //-------------------------------  Invalid clustering method
	    else {
	       error("Unknown cluster method");
	    }

	    // replace triggers
	    triggers = triggersClustered;

	    // report cluster numbers
	    if (debugLevel >= 1) {
	       // clusters[0][0].dump(cout);
	       clusters.status(cout);
	    }

	    ///////////////////////////////////////////////////////////////////
	    //                      write cluster triggers                   //
	    ///////////////////////////////////////////////////////////////////

	    if (outputFiles.enabled("CLUSTER")) {
	       wlog(debugLevel, 1, "  writing cluster triggers");

	       // write triggers or clusters
	       if (pars.writeClusters) {
		  clusters.writeEvents(outputFiles["CLUSTER"], 
				       pars.triggerFields,
				       pars.triggerFormat);
	       } 
	       else {
		  triggersClustered.writeEvents(outputFiles["CLUSTER"], 
						pars.triggerFields, 
						pars.triggerFormat);
	       }
	    }
	 } // end test for apply clustering

	 //////////////////////////////////////////////////////////////////////
	 //                      coincide triggers                           //
	 //////////////////////////////////////////////////////////////////////

#ifndef NO_COINCIDENCE
	 bool coincidentsFound = false;

	 // apply coincidence if (requested
	 if ((pars.coincidenceNumber > 1) && (numberOfChannels > 1)) {
	    wlog(debugLevel, 1, "  coinciding triggers");

	    // apply coincidence
	    wcoincide triggersCoincident(triggers, 
					 pars.coincidenceNumber, 
					 pars.durationInflation,
					 pars.bandwidthInflation, 
					 pars.maximumCoincidents, 
					 debugLevel);

	    // replace triggers
	    triggers = triggersCoincident;
  
	    // cull coincident clusters based on coincident triggers
	    if (exist("clusters","var")) {
	       wcoincidecluster clustersCoincident(triggersCoincident, 
						   clusters, debugLevel);
      
	       // replace clusters
	       clusters = clustersCoincident;
	    }

	    // note if (any coincidents were found
	    if (!triggersCoincident.empty()) {
	       coincidentsFound = true;
	    }

	    // report coincidence numbers
	    if (debugLevel >= 1) {
	       for (size_t chanNumber=0; chanNumber < triggersCoincident.size();
		    chanNumber++) {
		  cout <<  "    " << triggersCoincident[chanNumber].channelName 
		       << ": " << triggersCoincident[chanNumber].size() 
		       << " coincidents" << endl;
	       }
	    }

	    ///////////////////////////////////////////////////////////////////
	    //                write coincident triggers                      //
	    ///////////////////////////////////////////////////////////////////
  
	    if (outputFiles.enabled("COINCIDE")) {
	       wlog(debugLevel, 1, "  writing coincident triggers");

	       // write triggers or clusters
	       if (pars.applyClustering && pars.writeClusters) {
		  wwriteevents(clustersCoincident, outputFiles.COINCIDE, 
			       pars.triggerFields, pars.triggerFormat);
	       }
	       else {
		  wwriteevents(triggersCoincident, outputFiles.COINCIDE, 
			       pars.triggerFields, pars.triggerFormat);
	       }
	    }
	 }    // end test for apply coincidence

	 //////////////////////////////////////////////////////////////////////
	 //                       EVENT FOLLOWUP
	 //////////////////////////////////////////////////////////////////////

	 // apply followup if (requested and coincident tiles and clusters have 
	 // been found
	 if (coincidentsFound) {

	    ///////////////////////////////////////////////////////////////////
	    // create the output event structure
	    ///////////////////////////////////////////////////////////////////
	    event.id = "Discrete Q-transform event structure";
	    event.network = networkString;
	    event.blockStartTime = blockStartTime;
	    event.blockStopTime = blockStopTime;
	    event.livetime = pars.blockDuration - 2 * tiling.transientDuration;
  
	    event.outputFields = {"network", "blockStartTime", "blockStopTime", 
				  "livetime"};

	    ///////////////////////////////////////////////////////////////////
	    // find candidate event cluster
	    ///////////////////////////////////////////////////////////////////
	    wlog(debugLevel, 1, "  finding candidate cluster");

	    wfindcandidate candidateCluster(clustersCoincident, 
					    triggersClustered,
					    eventTime, pars, debugLevel);

	    event.time = candidateCluster.clusterTime;
	    event.frequency = candidateCluster.clusterFrequency;
	    event.duration = candidateCluster.clusterDuration;
	    event.bandwidth = candidateCluster.clusterBandwidth;

	    event.outputFields = {event.outputFields{:}, "time", "frequency", 
				  "duration", "bandwidth"};

	    ///////////////////////////////////////////////////////////////////
	    // FOLLOWUP ANALYSES
	    ///////////////////////////////////////////////////////////////////
#ifndef NO_FOLLOWUP
	    if (pars.bayesian || pars.xCoherentCheck || pars.maxEnt){
	       wlog(debugLevel, 1, "  Event followups:");

	       [event, skymap] = 
		  wfollowups(event, data, coefficients, tiling, 
			     blockStartTime, coordinate, candidateCluster, 
			     channelNames, pars, outputFiles, 
			     debugLevel);
	    }
#endif // !defined( NO_FOLLOWUP)

	    ///////////////////////////////////////////////////////////////////
	    // write event
	    ///////////////////////////////////////////////////////////////////
	    wlog(debugLevel, 1, "  writing event");

	    wwriteevents(event, outputFiles.EVENTS, event.outputFields, 
			 pars.triggerFormat);
	 } // end if (coincidentsFound)
#endif // !defined(NO_COINCIDENCE)
      } // if !channelNames.empty()
      blockNo++;

      /////////////////////////////////////////////////////////////////////////
      //                            write livetime                           //
      /////////////////////////////////////////////////////////////////////////
      wlog(debugLevel, 1, "  writing livetime");

      // analyzed channels string
      string analyzedChannels;
      for (size_t channelNumber=0; channelNumber<channelNames.size(); 
	   channelNumber++) {
	 if (channelNumber != 0) analyzedChannels += ",";
	 analyzedChannels += channelNames[channelNumber];
      }

      // open livetime file for appending
      ofstream liveStream(outputFiles["livetime"].path(0).c_str(), ios::app);

      // test for error
      if (!liveStream.is_open()) {
	 error("cannot open livetime file for writing");
      }

      // write block livetime to livetime file
      liveStream 
	 << setw(5) << blockNo << " " << fixed << setprecision(4)
	 << (blockStartTime + tiling.transientDuration()).getS() << "-"
	 << (blockStopTime  - tiling.transientDuration()).getS() << " "
	 << analyzedChannels << " " << cpuTime.get_time() << endl;

      // close livetime file
      liveStream.close();
   }


} // namespace wpipe
